package edu.mit.ll.group43.surfaceoptimization;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftAlphaComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator.ActionType;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
public class LookAheadSequencer extends Sequencer {
	
	public LookAheadSequencer(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
	}

	public int earliestReleaseAfter(Aircraft lastReleased, Aircraft leader, Aircraft follower) {
		
		int earlyReleaseBound = Math.max(follower.getSpotReadyTime(),
				lastReleased.getOptReleaseTime());
		
		// If there is a first aircraft (i.e. b isn't the first one in the
		// sequence), then set the separation between a and the current
		// aircraft and update the early release bound
		if (leader != null) {
			int earlySeparation = CALC.getSeparation(leader.getWeightClass(), ActionType.DEPARTURE,
					follower.getWeightClass(), ActionType.DEPARTURE);
			earlyReleaseBound = Math.max(earlyReleaseBound,
					leader.getRunwayReadyTime() + earlySeparation - follower.getUnimpTaxiTime());
		}
		
		return earlyReleaseBound;
	}
	
	public boolean canFitBefore(int earlyReleaseBound, Aircraft current, Aircraft lastTakeOff) {
		int lateSeparation = CALC.getSeparation(current.getWeightClass(), ActionType.DEPARTURE,
				lastTakeOff.getWeightClass(), ActionType.DEPARTURE);
		return earlyReleaseBound + current.getUnimpTaxiTime() + lateSeparation
				<= lastTakeOff.getRunwayReadyTime();
	}
	
	public int earliestReleaseBetween(Aircraft lastReleased, Aircraft a, Aircraft b, Aircraft current) {
		
		int earlyReleaseBound = earliestReleaseAfter(lastReleased, a, current);
		
		// If this aircraft can fit between a and b (release time + taxi time + separation
		// <= b's arrival at the runway) then return that release time
		if (canFitBefore(earlyReleaseBound, current, b)) {
			return earlyReleaseBound;
		}
		
		// Otherwise, it can't be done
		return -1;
	}
	
	/**
	 * Best spot release time where follower is released and takes off after leader.
	 * 
	 * @param takeOffSequence
	 * @param leader
	 * @param follower
	 * @return
	 */
	public int bestStart(ArrayList<Aircraft> releaseSequence, Aircraft leader, Aircraft follower) {
		
		ArrayList<Aircraft> takeOffSequence = new ArrayList<Aircraft>(releaseSequence);
		Collections.sort(takeOffSequence, AircraftAlphaComparator.getInstance());
		Aircraft a = null;
		Aircraft b = takeOffSequence.get(takeOffSequence.size()-1);
		if (takeOffSequence.size() > 1) {
			a = takeOffSequence.get(takeOffSequence.size()-2);
		}
		int onegap = earliestReleaseBetween(leader, a, b, follower);
		if (onegap > -1) {
			return onegap;
		}
		
		// Earliest = max{delta(i,j) - (uj - ui) + soi, soj}
		int earliest = b.getOptReleaseTime() + CALC.getSeparation(b.getWeightClass(),
				ActionType.DEPARTURE, follower.getWeightClass(), ActionType.DEPARTURE);
		earliest = earliest - (follower.getUnimpTaxiTime() - b.getUnimpTaxiTime());
		earliest = Math.max(earliest, follower.getSpotReadyTime());
		earliest = Math.max(earliest, leader.getOptReleaseTime());
		
		return earliest;
	}
	
	public int totalDelay(ArrayList<Aircraft> releaseSequence, Aircraft lastRelease,
			Aircraft leader, Aircraft follower) {
		int delay = bestStart(releaseSequence, lastRelease, leader) - leader.getSpotReadyTime();
		Aircraft firstNextClone = leader.clone();
		firstNextClone.setOptReleaseTime(delay + leader.getSpotReadyTime());
		releaseSequence.add(firstNextClone);
		delay += bestStart(releaseSequence, firstNextClone, follower) - follower.getSpotReadyTime();
		releaseSequence.remove(firstNextClone);
		
		return delay;
	}
	
	public void placeAircraft(ArrayList<Aircraft> releaseSequence, Aircraft prev, Aircraft first) {
		first.setOptReleaseTime(bestStart(releaseSequence, prev, first));
		releaseSequence.add(first);
	}
	
	public void placeAircraft(ArrayList<Aircraft> releaseSequence, Aircraft prev, Aircraft first, Aircraft second) {
		first.setOptReleaseTime(bestStart(releaseSequence, prev, first));
		releaseSequence.add(first);
		if (second != null) {
			second.setOptReleaseTime(bestStart(releaseSequence, first, second));
			releaseSequence.add(second);
		}
	}
	
	public void updateSequence(ArrayList<Aircraft> seq, Aircraft first, Aircraft second) {
		seq.remove(first);
		seq.remove(second);
	}
	
	@Override
	public ArrayList<Aircraft> sequence(ArrayList<Aircraft> seq) {
		
		ArrayList<Aircraft> releaseSequence = new ArrayList<Aircraft>();
		
		// Send the first aircraft that is ready first
		Aircraft first = seq.remove(0);
		first.setOptReleaseTime(first.getSpotReadyTime());
		releaseSequence.add(first);
		
		// Iteratively look at the next k+1 aircraft and determine which
		// is the best to be released next, and at what time, but only if
		// CPS is not broken
		while (seq.size() > 0) {
			
			// Get the aircraft last released and the next two to place
			Aircraft lastRelease = releaseSequence.get(releaseSequence.size()-1);
			List<Aircraft> lookAhead = new ArrayList<Aircraft>(seq.subList(0, Math.min(2, seq.size())));
			Aircraft firstNext = lookAhead.get(0);
			Aircraft secondNext = null;
			
			if (lookAhead.size() > 1) {
				
				// Calculate the delay of sending the first one ready first
				secondNext = lookAhead.get(1);
				int firstDelay = totalDelay(releaseSequence, lastRelease, firstNext, secondNext);
				
				// Calculate the delay of sending the second one ready first
				int secondDelay = totalDelay(releaseSequence, lastRelease, secondNext, firstNext);
				
				// Determine if one or the other must go first to obey CPS
				boolean firstMust = releaseSequence.size() + 2 - firstNext.getFcfsIndex() > CPS;
				boolean secondMust = releaseSequence.size() + 2 - secondNext.getFcfsIndex() > CPS;
				
				// Determine which to send first
				if (firstMust || firstDelay < secondDelay && !secondMust) {
					placeAircraft(releaseSequence, lastRelease, firstNext);
					seq.remove(firstNext);
				} else {
					placeAircraft(releaseSequence, lastRelease, secondNext);
					seq.remove(secondNext);
				}
				
				// Remove both from the FCFS sequence so they are not
				// considered again
				//updateSequence(seq, firstNext, secondNext);
			} else {
				placeAircraft(releaseSequence, lastRelease, firstNext);
				seq.remove(firstNext);
			}
		}
		
		// Return the release sequence
		return releaseSequence;
	}

}