package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;
import java.util.Collections;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftAlphaComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftSpotReleaseComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator.ActionType;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.SmartCPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.validator.CPSValidator;

/**
 * Solves for the shortest path through the CPS network by implementing
 * the All Gaps Force heuristic. This algorithm considers all gaps in
 * the projected takeoff sequence that the aircraft currently being
 * scheduled could possibly fit into, with respect to its spot ready time
 * and unimpeded taxi time. These gaps may be "forced" to fit the current
 * aircraft in them, which involves rescheduling the last aircraft of
 * the gap so that no runway delay is incurred. The spot release time for
 * the current aircraft that minimizes the additional delay by forcing a
 * gap or scheduling it to take off last is solved for.
 * 
 * @author William Hawkins
 *
 */
public class AllGapsForceSpotCostStrategy extends OneGapForceSpotCostStrategy {

	protected ArrayList<Aircraft> bestSequence;
	protected int bestTotalDelay;
	protected int bestReleaseTime;
	
	protected ArrayList<Aircraft> currBestSequence;
	protected int currBestTotalDelay;
	protected int currBestReleaseTime;
	
	public AllGapsForceSpotCostStrategy(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
		clearBestFields();
		clearCurrFields();
	}
	
	/**
	 * Sets the optimal values for the current node and its last aircraft,
	 * and saves the spot release sequence that corresponds to those values
	 * to the node.
	 * 
	 * @param edge
	 * @param distance
	 */
	@Override
	public void setOptimalValues(NetworkEdge edge, int distance) {
		SmartCPSNetworkNode node = (SmartCPSNetworkNode) edge.getTarget();
		Aircraft follower = node.getLastAircraft();
		
		// Set the optimal distance, prev node, and spot release time
		node.setOptDistance(distance);
		node.setOptPrevNode(edge);
		follower.setOptReleaseTime(bestReleaseTime);
		
		// Set this node's release sequence
		setBestReleaseSequences(node);
		
		// Reset variables
		clearBestFields();
	}
	
	/**
	 * Sets the best release sequence of the given node.
	 * 
	 * @param node
	 */
	public void setBestReleaseSequences(SmartCPSNetworkNode node) {
		node.setReleaseSequence(bestSequence);
	}
	
	/**
	 * Initializes the fields that hold values associated with the current
	 * aircraft being placed.
	 * 
	 */
	public void clearBestFields() {
		bestSequence = null;
		bestTotalDelay = -1;
		bestReleaseTime = -1;
	}
	
	/**
	 * Initializes the fields that hold values associated with the current
	 * predecessor node.
	 * 
	 */
	public void clearCurrFields() {
		currBestTotalDelay = -1;
		currBestReleaseTime = -1;
		currBestSequence = null;
	}
	
	/**
	 * Returns an ArrayList of cloned aircraft from the given sequence, sorted
	 * in order of their optimal spot release time.
	 * 
	 * @param sequence
	 * @return
	 */
	public ArrayList<Aircraft> sortedReleases(ArrayList<Aircraft> sequence) {
		ArrayList<Aircraft> releaseSequence = new ArrayList<Aircraft>();
		for (Aircraft ac : sequence) {
			releaseSequence.add(ac.clone());
		}
		Collections.sort(releaseSequence, AircraftSpotReleaseComparator.getInstance());
		return releaseSequence;
	}
	
	/**
	 * Modifies the given takeoff sequence to not have any conflicts at the runway,
	 * and returns the delay caused by this. Only looks at the aircraft between and
	 * including the given start and end indexes.
	 * 
	 * @param takeOffSequence
	 * @param startIndex
	 * @param endIndex
	 * @return
	 */
	public int adjustTakeOffs(ArrayList<Aircraft> takeOffSequence, int startIndex, int endIndex) {
		
		int adjustmentDelay = 0;
		
		if (startIndex < 1) {
			startIndex = 1;
		}
		
		for (int i = startIndex; i < takeOffSequence.size(); i ++) {
			
			// Calculate how much the second aircraft needs to move
			Aircraft first = takeOffSequence.get(i-1);
			Aircraft second = takeOffSequence.get(i);
			int takeOffDiff = first.getOptReleaseTime() + first.getUnimpTaxiTime() +
					CALC.getSeparation(first.getWeightClass(), ActionType.DEPARTURE,
							second.getWeightClass(), ActionType.DEPARTURE) - second.getRunwayReadyTime();
			
			// If there is no need to move the second aircraft, there will be
			// no more conflicts at the runway
			if (takeOffDiff <= 0) {
				takeOffDiff = 0;
				if (i > endIndex) {
					break;
				}
			}
			
			// Update when the second aircraft should be released to avoid
			// conflict at the runway, and update the total amount of delay
			// that shifting previously scheduled aircraft has caused
			second.setOptReleaseTime(second.getOptReleaseTime() + takeOffDiff);
			adjustmentDelay += takeOffDiff;
		}
		
		return adjustmentDelay;
	}
	
	/**
	 * Considers what would happen if the current aircraft were scheduled to take
	 * off between aircraft a and b. Updates the best sequence found so far if
	 * the resulting sequence is valid and incurs less delay than the one currently
	 * saved as the best sequence.
	 * 
	 * The takeoff sequence should be cloned, and a and b should be clones from
	 * that sequence. The presence of the sequence allows for other aircraft
	 * to be adjusted if necessary based on b's movement. The current aircraft
	 * should also be a clone.
	 * 
	 * @param a
	 * @param b
	 * @param current
	 * @param takeOffSequence
	 */
	public void computeValues(Aircraft a, Aircraft b, Aircraft current,
			ArrayList<Aircraft> takeOffSequence) {
		
		Collections.sort(takeOffSequence, AircraftSpotReleaseComparator.getInstance());
		int earliestRelease = earliestReleaseAfter(safePrevTimeBound(takeOffSequence, current), a, current);
		
		// Determine how this affects b. Aircraft b will need to make its
		// optimal spot release time later than it is by the number of
		// seconds of overlap between follower's takeoff and wake vortex
		// separation and b's runway ready time. This difference will add delay.
		int diff = earliestRelease + current.getUnimpTaxiTime() + 
				CALC.getSeparation(current.getWeightClass(), ActionType.DEPARTURE, 
						b.getWeightClass(), ActionType.DEPARTURE) - b.getRunwayReadyTime();
		
		// The difference should be at least zero, diff could be negative at this
		// point if there is no conflict, which could schedule b to leave the spot
		// before it is ready
		if (diff < 0) {
			diff = 0;
		}
		
		// Initialize the total delay due to rescheduling of previously scheduled
		// aircraft to how much longer b is held back
		int diffDelay = diff;
		
		// If they do overlap, hold b's optimal release time back by that amount
		// so that the effects can be measured below when the list is sorted
		b.setOptReleaseTime(b.getOptReleaseTime() + diff);
		current.setOptReleaseTime(earliestRelease);
		takeOffSequence.add(current);
		Collections.sort(takeOffSequence, AircraftAlphaComparator.getInstance());
		
		// Adjust the other aircraft in the sequence as necessary
		diffDelay += adjustTakeOffs(takeOffSequence, takeOffSequence.indexOf(current),
				takeOffSequence.indexOf(b));
		
		// The total delay for placing current between a and b is the diff delay
		// plus the delay for current
		int totalDelay = diffDelay + current.getOptReleaseTime() - current.getSpotReadyTime();
		
		// Determine if this is the best valid spot release time for current so far
		updateCurrValues(takeOffSequence, totalDelay, earliestRelease);
	}
	
	/**
	 * Updates the best sequence found so far for this node and its associated values
	 * if the given sequence is valid and incurs less delay than the one currently
	 * saved, or if there is no currently saved best sequence for this node.
	 * 
	 * @param sequence
	 * @param totalDelay
	 * @param release
	 */
	public void updateCurrValues(ArrayList<Aircraft> sequence, int totalDelay,
			int release) {
		if (totalDelay < currBestTotalDelay || currBestTotalDelay == -1) {
			Collections.sort(sequence, AircraftSpotReleaseComparator.getInstance());
			//TODO Figure out a way to keep track of the earliest one shifted to limit
			// how much of the sequence must be validated
			if (CPSValidator.validate(sequence, CPS)) {
				currBestTotalDelay = totalDelay;
				currBestReleaseTime = release;
				currBestSequence = sequence;
			}
		}
	}
	
	/**
	 * Returns the index of the earliest aircraft it should look at
	 * "forcing" (i.e. the earliest "b").
	 * 
	 * @param releaseSequence
	 * @param current
	 * @param limit
	 * @return
	 */
	public int earliestGap(ArrayList<Aircraft> releaseSequence, Aircraft current, int limit) {
		
		int safePrevTimeBound = safePrevTimeBound(releaseSequence, current);
		ArrayList<Aircraft> takeOffSequence = sortedAlphasSoFar(releaseSequence);
		
		for (int i = takeOffSequence.size()-1; i >= 0; i --) {
			Aircraft a = null;
			if (i > 0) {
				a = takeOffSequence.get(i-1);
			}
			Aircraft b = takeOffSequence.get(i);
			int earliestRelease = earliestReleaseAfter(safePrevTimeBound, a, current);
			int diff = earliestRelease + current.getUnimpTaxiTime() + 
					CALC.getSeparation(current.getWeightClass(), ActionType.DEPARTURE,
							b.getWeightClass(), ActionType.DEPARTURE) - b.getRunwayReadyTime();
			
			if (diff > limit) {
				return i + 1;
			}
		}
		
		return 0;
	}
	
	/**
	 * Determines if the given "normal" sequence, which schedules the current
	 * aircraft to take off last, is better than any of the sequences that
	 * force gaps. If it is, then the best sequence for the current previous
	 * node being looked at is updated, along with its associated values.
	 * 
	 * @param normalSequence
	 * @param normalDelay
	 * @param normalRelease
	 */
	public void compareNormalApproach(ArrayList<Aircraft> normalSequence, 
			int normalDelay, int normalRelease) {
		if (normalDelay < currBestTotalDelay || currBestTotalDelay == -1) {
			currBestSequence = normalSequence;
			currBestTotalDelay = normalDelay;
			currBestReleaseTime = normalRelease;
		}
	}
	
	/**
	 * Solves for the best schedule where the current aircraft's optimal
	 * spot release time is the most recently calculated, when following
	 * the previous node currently being considered.
	 * 
	 * @param node
	 * @param current
	 */
	public void scheduleCurrentAircraft(SmartCPSNetworkNode node, Aircraft current) {
		ArrayList<Aircraft> releases = node.getReleaseSequence();
		updateScheduleForSequence(releases, current);
	}
	
	/**
	 * Calculates the "predecessor bound" that should be used when scheduling the
	 * current aircraft to take off last in the current projected takeoff sequence.
	 * 
	 * @param releases
	 * @param current
	 * @return
	 */
	public int normalSafePrevTimeBound(ArrayList<Aircraft> releases, Aircraft current) {
		return CostStrategyUtils.safePrevTimeBound(releases, current, CPS);
	}
	
	/**
	 * Solves for the best schedule where the current aircraft's optimal
	 * spot release time is the most recently calculated, when following
	 * the previous node currently being considered, when basing this
	 * optimal schedule on the one given.
	 * 
	 * @param releases
	 * @param current
	 */
	public void updateScheduleForSequence(ArrayList<Aircraft> releases, Aircraft current) {
		
		// Get what would happen by scheduling current to take off last
		Aircraft lastTakeOff = CostStrategyUtils.lastTakeOff(releases);
		int normalRelease = earliestRelease(lastTakeOff, current, normalSafePrevTimeBound(releases, current));
		int normalDelay = normalRelease - current.getSpotReadyTime();
		
		ArrayList<Aircraft> normalSequence = CostStrategyUtils.cloneSequence(releases);
		current.setOptReleaseTime(normalRelease);
		normalSequence.add(current);
		Collections.sort(normalSequence, AircraftSpotReleaseComparator.getInstance());
		
		// Compute values for all gaps, which will set the fields to the best
		// solution found among them
		for (int i = earliestGap(releases, current, normalDelay); i < releases.size(); i ++) {
			ArrayList<Aircraft> takeOffSequence = sortedAlphasSoFar(releases);
			Aircraft a = null;
			Aircraft b = takeOffSequence.get(i);
			if (i > 0) {
				a = takeOffSequence.get(i-1);
			}
			computeValues(a, b, current.clone(), takeOffSequence);
		}
		
		compareNormalApproach(normalSequence, normalDelay, normalRelease);
	}
	
	/**
	 * Checks the best values found for the current previous node being looked at
	 * against the best values found among all previous nodes for the current
	 * node whose last aircraft is being scheduled. Updates the best values if
	 * the values for the current previous node are better.
	 * 
	 */
	public void updateBestValues() {
		if (currBestTotalDelay < bestTotalDelay || bestTotalDelay == -1) {
			bestSequence = currBestSequence;
			bestTotalDelay = currBestTotalDelay;
			bestReleaseTime = currBestReleaseTime;
		}
	}
	
	/**
	 * Returns the additional delay associated with scheduling the current aircraft
	 * according to the All Gaps Force heuristic algorithm for the given edge
	 * in the network.
	 * 
	 * @param edge
	 */
	@Override
	public int getEdgeWeight(NetworkEdge edge) {
		
		// Initialize fields specific to saving info for the current prev node
		clearCurrFields();
		
		// References to source node and current aircraft
		SmartCPSNetworkNode optPrevNode = (SmartCPSNetworkNode) edge.getSource();
		Aircraft current = ((CPSNetworkNode) edge.getTarget()).getLastAircraft().clone();
		
		// Copy the schedule from the previous node and solve for the best schedule
		// that includes the current aircraft
		scheduleCurrentAircraft(optPrevNode, current);
		
		// Determine if this solution is better than the best found so far
		updateBestValues();
		
		// Return the best additional delay found from this prev node, regardless
		// of whether it is better than the best found so far
		return currBestTotalDelay;
	}

}
