package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;
import java.util.TreeSet;

import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;

/**
 * This abstract class is extended by classes that act as edge
 * cost strategies for the DPSequencer. These classes are
 * responsible for helping to solve the shortest path in the
 * generated CPS network by correctly initializing nodes in
 * the first stage, calculating what a given edge weight should
 * be, and properly updating the values of a given node to ensure
 * an optimal solution is achieved.
 * 
 * @author William Hawkins
 *
 */
public abstract class CostStrategy {

	protected int CPS;
	protected MinimumSeparationCalculator<? extends WeightClass> CALC;
	
	public CostStrategy(int cps, MinimumSeparationCalculator<? extends WeightClass> calc) {
		CPS = cps;
		CALC = calc;
	}
	
	/**
	 * Initializes the first stage of the CPS network so that subsequent
	 * stages may be solved for.
	 * 
	 * @param network
	 */
	public abstract void initializeFirstStage(ArrayList<TreeSet<CPSNetworkNode>> network);
	
	/**
	 * Sets the optimal values of the node currently being solved, as well
	 * as for its last aircraft.
	 * 
	 * @param edge
	 * @param distance
	 */
	public abstract void setOptimalValues(NetworkEdge edge, int distance);
	
	/**
	 * Returns the additional delay incurred when the optimal spot release
	 * time of the last aircraft of the target node of the edge is calculated
	 * immediately after that of the source node of the edge.
	 * 
	 * @param edge
	 * @return
	 */
	public abstract int getEdgeWeight(NetworkEdge edge);
}
