package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.measurement.CPSDeviationMeasurer;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.LenientSmartCPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.SmartCPSNetworkNode;

/**
 * Cost strategy for the Lenient All Gaps Force algorithm. This algorithm
 * operates in the same way as All Gaps Force, but as it solves the network,
 * it has each node save two sequences that are invalid according to CPS in
 * addition to the one valid sequence each node saves. These invalid
 * sequences are:
 * - The sequence that comes closest to being valid among those that have the
 * lowest total delays
 * - The sequence with the lowest total delay among those that are the closest
 * to being valid sequences
 * 
 * Better results may be obtained by considering intermediate sequences that
 * are invalid that could be corrected due to forcing rather than only
 * considering sequences that are valid throughout the optimization.
 * 
 * @author William Hawkins
 *
 */
public class LenientAllGapsForceSpotCostStrategy extends
		AllGapsForceSpotCostStrategy {

	protected ArrayList<Aircraft> bestCorrectInvalidSequence;
	protected int bestCorrectInvalidTotalDelay;
	protected int bestCorrectInvalidReleaseTime;
	protected int bestCorrectInvalidDeviation;
	
	protected ArrayList<Aircraft> currBestCorrectInvalidSequence;
	protected int currBestCorrectInvalidTotalDelay;
	protected int currBestCorrectInvalidReleaseTime;
	protected int currBestCorrectInvalidDeviation;
	
	protected ArrayList<Aircraft> bestEfficientInvalidSequence;
	protected int bestEfficientInvalidTotalDelay;
	protected int bestEfficientInvalidReleaseTime;
	protected int bestEfficientInvalidDeviation;
	
	protected ArrayList<Aircraft> currBestEfficientInvalidSequence;
	protected int currBestEfficientInvalidTotalDelay;
	protected int currBestEfficientInvalidReleaseTime;
	protected int currBestEfficientInvalidDeviation;
	
	public LenientAllGapsForceSpotCostStrategy(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
	}
	
	/**
	 * Resets the fields that keep track of the best sequences and their
	 * associated values that include the current aircraft being scheduled.
	 * 
	 */
	@Override
	public void clearBestFields() {
		super.clearBestFields();
		
		bestCorrectInvalidSequence = null;
		bestCorrectInvalidTotalDelay = -1;
		bestCorrectInvalidReleaseTime = -1;
		bestCorrectInvalidDeviation = -1;
		
		bestEfficientInvalidSequence = null;
		bestEfficientInvalidTotalDelay = -1;
		bestEfficientInvalidReleaseTime = -1;
		bestEfficientInvalidDeviation = -1;
	}
	
	/**
	 * Resets the fields that keep track of the best sequences and their
	 * associated values that include the current aircraft being scheduled,
	 * given the previous node currently under consideration.
	 * 
	 */
	@Override
	public void clearCurrFields() {
		super.clearCurrFields();
		
		currBestCorrectInvalidSequence = null;
		currBestCorrectInvalidTotalDelay = -1;
		currBestCorrectInvalidReleaseTime = -1;
		currBestCorrectInvalidDeviation = -1;
		
		currBestEfficientInvalidSequence = null;
		currBestEfficientInvalidTotalDelay = -1;
		currBestEfficientInvalidReleaseTime = -1;
		currBestEfficientInvalidDeviation = -1;
	}
	
	/**
	 * Sets the best valid and invalid sequences for the node of the
	 * aircraft currently being scheduled.
	 * 
	 * @param node
	 */
	@Override
	public void setBestReleaseSequences(SmartCPSNetworkNode node) {
		super.setBestReleaseSequences(node);
		((LenientSmartCPSNetworkNode) node).setMostCorrectInvalidSequence(bestCorrectInvalidSequence);
		((LenientSmartCPSNetworkNode) node).setMostEfficientInvalidSequence(bestEfficientInvalidSequence);
	}
	
	/**
	 * Returns zero as the "predecessor bound" since setting it the same way that
	 * other forcing algorithms set it would limit the invalid sequences that it
	 * considers.
	 * 
	 * @param optPrevNode
	 * @param current
	 */
	@Override
	public int safePrevTimeBound(CPSNetworkNode optPrevNode, Aircraft current) {
		return 0;
	}
	
	/**
	 * Schedules the current aircraft and determines the best sequences that include it
	 * according to the best sequences saved in the previous node currently being considered.
	 * 
	 * @param node
	 * @param current
	 */
	@Override
	public void scheduleCurrentAircraft(SmartCPSNetworkNode node, Aircraft current) {
		super.scheduleCurrentAircraft(node, current.clone());
		
		LenientSmartCPSNetworkNode optPrevNode = (LenientSmartCPSNetworkNode) node;
		ArrayList<Aircraft> correctInvalidSequence = optPrevNode.getMostCorrectInvalidSequence();
		ArrayList<Aircraft> efficientInvalidSequence = optPrevNode.getMostEfficientInvalidSequence();
		
		if (correctInvalidSequence != null) {
			updateScheduleForSequence(correctInvalidSequence, current.clone());
		}
		
		if (efficientInvalidSequence != null) {
			updateScheduleForSequence(efficientInvalidSequence, current.clone());
		}
	}
	
	/**
	 * Updates the best recorded values for the aircraft currently being scheduled
	 * based on the best values recorded for the previous node currently under
	 * consideration.
	 * 
	 */
	@Override
	public void updateBestValues() {
		super.updateBestValues();
		
		if (currBestCorrectInvalidDeviation < bestCorrectInvalidDeviation ||
				currBestCorrectInvalidDeviation == bestCorrectInvalidDeviation &&
				currBestCorrectInvalidTotalDelay < bestCorrectInvalidTotalDelay
				|| bestCorrectInvalidTotalDelay == -1) {
			bestCorrectInvalidSequence = currBestCorrectInvalidSequence;
			bestCorrectInvalidTotalDelay = currBestCorrectInvalidTotalDelay;
			bestCorrectInvalidReleaseTime = currBestCorrectInvalidReleaseTime;
			bestCorrectInvalidDeviation = currBestCorrectInvalidDeviation;
		}
		
		if (currBestEfficientInvalidTotalDelay < bestEfficientInvalidTotalDelay ||
				currBestEfficientInvalidTotalDelay == bestEfficientInvalidTotalDelay &&
				currBestEfficientInvalidDeviation < bestEfficientInvalidDeviation
				|| bestEfficientInvalidTotalDelay == -1) {
			bestEfficientInvalidSequence = currBestEfficientInvalidSequence;
			bestEfficientInvalidTotalDelay = currBestEfficientInvalidTotalDelay;
			bestEfficientInvalidReleaseTime = currBestEfficientInvalidReleaseTime;
			bestEfficientInvalidDeviation = currBestEfficientInvalidDeviation;
		}
	}
	
	/**
	 * Tests the given sequence and its associated values against the best
	 * recorded values for the previous node currently under consideration.
	 * 
	 * @param sequence
	 * @param totalDelay
	 * @param release
	 */
	@Override
	public void updateCurrValues(ArrayList<Aircraft> sequence, int totalDelay,
			int release) {
		
		int deviation = CPSDeviationMeasurer.measure(sequence, CPS);
		if (deviation == 0) {
			super.compareNormalApproach(sequence, totalDelay, release);
		} else {
			if (deviation < currBestCorrectInvalidDeviation ||
					deviation == currBestCorrectInvalidDeviation &&
					totalDelay < currBestCorrectInvalidTotalDelay
					|| currBestCorrectInvalidTotalDelay == -1) {
				currBestCorrectInvalidSequence = sequence;
				currBestCorrectInvalidTotalDelay = totalDelay;
				currBestCorrectInvalidReleaseTime = release;
				currBestCorrectInvalidDeviation = deviation;
			}
			
			if (totalDelay < currBestEfficientInvalidTotalDelay ||
					totalDelay == currBestEfficientInvalidTotalDelay &&
					deviation < currBestEfficientInvalidDeviation
					|| currBestEfficientInvalidTotalDelay == -1) {
				currBestEfficientInvalidSequence = sequence;
				currBestEfficientInvalidTotalDelay = totalDelay;
				currBestEfficientInvalidReleaseTime = release;
				currBestEfficientInvalidDeviation = deviation;
			}
		}
	}
	
	/**
	 * Compares scheduling the current aircraft to take off last to the
	 * sequences that involve forcing.
	 * 
	 * @param normalSequence
	 * @param normalDelay
	 * @param normalRelease
	 */
	@Override
	public void compareNormalApproach(ArrayList<Aircraft> normalSequence, 
			int normalDelay, int normalRelease) {
		updateCurrValues(normalSequence, normalDelay, normalRelease);
	}
}
