package edu.mit.ll.group43.surfaceoptimization.dp.factory.netgen;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeSet;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.edge.EdgeValueStrategy;
import edu.mit.ll.group43.surfaceoptimization.dp.factory.node.CPSNetworkNodeFactory;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.Subsequence;
import edu.mit.ll.group43.surfaceoptimization.network.dp.SubsequenceGeneratorNode;

public class GeneratorNetworkCPSNetworkGenerator extends
		CPSNetworkGenerator {

	public GeneratorNetworkCPSNetworkGenerator(int cps,
			CPSNetworkNodeFactory networkNodeFactory, EdgeValueStrategy edgeValueStrategy) {
		super(cps, networkNodeFactory, edgeValueStrategy);
	}
	
	/**
	 * Generates the subsequence generator network that is used to generate the nodes for
	 * the CPS network.
	 * 
	 * @param seqLength
	 * @return
	 */
	public ArrayList<ArrayList<SubsequenceGeneratorNode>> generateSubsequenceGeneratorNetwork(int seqLength) {
		
		// Create the nodes
		ArrayList<ArrayList<SubsequenceGeneratorNode>> stages = generateSubsequenceGeneratorNodes(seqLength);
		
		// Add the edges and relationships and return it
		return generateSubsequenceGeneratorNodeRelationships(stages);
	}
	
	/**
	 * Generates the nodes for the subsequence generator network that is used to generate
	 * the nodes for the CPS network.
	 * 
	 * @param seqLength
	 * @return
	 */
	public ArrayList<ArrayList<SubsequenceGeneratorNode>> generateSubsequenceGeneratorNodes(int seqLength) {
		
		// Initialize what is assignable where (last position of each subsequence)
		HashMap<Integer, ArrayList<Integer>> validLocs = getValidLocs(seqLength);
		
		// Create the subsequence generator nodes
		ArrayList<ArrayList<SubsequenceGeneratorNode>> stages = new ArrayList<ArrayList<SubsequenceGeneratorNode>>();
		for (int stage = 1; stage <= seqLength; stage ++) {
			ArrayList<SubsequenceGeneratorNode> nodes = new ArrayList<SubsequenceGeneratorNode>();
			for (int ac : validLocs.get(stage)) {
				nodes.add(new SubsequenceGeneratorNode(ac));
			}
			stages.add(nodes);
		}
		
		return stages;
	}
	
	/**
	 * Adds correct relationships to the network that is used to create the subsequences stored
	 * in the nodes of the CPS network.
	 * 
	 * @param stages
	 * @return
	 */
	public ArrayList<ArrayList<SubsequenceGeneratorNode>> generateSubsequenceGeneratorNodeRelationships
		(ArrayList<ArrayList<SubsequenceGeneratorNode>> stages) {
		
		for (int i = 0; i < stages.size()-1; i ++) {
			ArrayList<SubsequenceGeneratorNode> stage = stages.get(i);
			ArrayList<SubsequenceGeneratorNode> nextStage = stages.get(i+1);
			for (SubsequenceGeneratorNode j : stage) {
				for (SubsequenceGeneratorNode k : nextStage) {
					j.addNextNode(new NetworkEdge(j, k, 0));
					k.addPrevNode(new NetworkEdge(j, k, 0));
				}
			}
		}
		
		return stages;
	}
	
	/**
	 * Goes through the subsequence generator network, computing all valid paths
	 * of length at most 2k + 1
	 * 
	 * @param stages
	 * @return
	 */
	public ArrayList<ArrayList<SubsequenceGeneratorNode>> generateValidPaths
		(ArrayList<ArrayList<SubsequenceGeneratorNode>> stages) {
		
		// Initialize the paths of the first stage
		for (int i = 0; i < stages.get(0).size(); i ++) {
			stages.get(0).get(i).addValidPath(new Subsequence<Integer>(), CPS);
		}
		
		// Create paths for all other stages
		for (int stage = 1; stage < stages.size(); stage ++) {
			for (SubsequenceGeneratorNode node : stages.get(stage)) {
				for (NetworkEdge edge : node.getPrevNodes()) {
					for (Subsequence<Integer> path : ((SubsequenceGeneratorNode)edge.getSource()).getValidPaths()) {
						// Automatically validates the path and takes only the
						// appropriate length of the path
						node.addValidPath(path, CPS);
					}
				}
			}
		}
		
		return stages;
	}
	
	@Override
	public ArrayList<TreeSet<CPSNetworkNode>> generateCPSNetworkNodes(ArrayList<Aircraft> fcfs) {
		// Generate the network that will generate the CPS network
		ArrayList<ArrayList<SubsequenceGeneratorNode>> stages =  
			generateSubsequenceGeneratorNetwork(fcfs.size());
		
		// Fill the network with valid CPS path information
		stages = generateValidPaths(stages);
		
		// Create the CPS network nodes
		ArrayList<TreeSet<CPSNetworkNode>> cpsStages = new ArrayList<TreeSet<CPSNetworkNode>>();
		for (ArrayList<SubsequenceGeneratorNode> stage : stages) {
			TreeSet<CPSNetworkNode> cpsStage = new TreeSet<CPSNetworkNode>();
			for (SubsequenceGeneratorNode node : stage) {
				for (Subsequence<Integer> seq : node.getValidPaths()) {
					Subsequence<Aircraft> ac = new Subsequence<Aircraft>();
					for (int i = 0; i < seq.size()-1; i ++) {
						// Add references of all aircraft in the sequence except the
						// last one
						ac.add(fcfs.get(seq.get(i)-1));
					}
					// Add a clone of the last aircraft to avoid logical errors during
					// the running of the algorithm arising from the same object
					// reference being in multiple nodes
					ac.add(fcfs.get(seq.get(seq.size()-1)-1).clone());
					cpsStage.add(cpsNetworkNodeFactory.getCPSNetworkNode(ac));
				}
			}
			cpsStages.add(cpsStage);
		}
		
		return cpsStages;
	}

}
